/*
  Babix - An educational little preemptive multitask kernel for Arduino Due.
  Francois Pessaux 04/2015.
  Maxime Ayrault 06/2015.
  This code can be freely distributed.
*/

#include <Arduino.h>
#include "kernel.h"
#include "context.h"
#include "process.h"
#include "queue.h"
#include "mutex.h"


/* The following variables are made global to avoid having them on the context
   switcher stack. This avoids the stack frame being modified and simplifies
   the context switch. */
/** Process to make running. */
static struct process_t *incoming_proc ;
static proc_id_t incoming_proc_id ;       /** Its id. */

/** Process to stop running. */
static struct process_t *outgoing_proc ;
static proc_id_t outgoing_proc_id ;       /** Its id. */


#ifdef __cplusplus
extern "C" { /* Start of extern "C" for svcHook (). */
#endif

/** The pendSVC instruction (system call) handler.
    Overrides the weak symbol of the Arduino "standard library".
    ATTENTION: must be extern "C" otherwise ignored by the linker, hence
    never called (no effective overriding).

    ATTENTION: Must be "naked" to prevent GCC from making its own stack frame.
    Since all the registers will be trashed by this function when restoring a
    context, there is no need for any registers backup. Even more, this would
    change our expected stack frame !

    The SAM3X8E automatically saves r0, r1, r2, r3, r12, the return address,
    psr and lr at exception entry. See SAM3x8E datasheet 16.6.7.5 page 85.
*/
/*
  Note on hardware exclusive access to memory
  -------------------------------------------
  Why we don't use a CLREX in the context switch routine with our LDREX and
  STREX in order to clear a possibly pending exclusive access request ?

  -> What is CLREX

  CLREX is an instruction clearing any possibly pending exclusive access,
  causing any STREX to fail, hence causing an exclusive store to fail
  (c.f. datasheet section 12.12.9 page 114).

  -> Why we need it generally ?

  Here is an example to explain why CLREX *could* be useful.

  Let's have 2 variables A and B and 3 processes P1, P2 and P3.

  P1 does a LDREX on B and set the exclusive tag.
  P1 is switched, and some things are done.
  P3 does a LDREX on B, and set the tag that is already set,
    and does a STREX on B that succeeds because the tag is set.
    The exclusive access tag is now unset.
  P3 is switched.
  P2 does a LDREX on A which sets the tag.
  P2 is switched.
  P1 comes and does STREX on B wich succeeds because P2 has set the tag.

  So the memory is corrupt.
  That is why *could* need a CLREX at the beginning of context switch, in order
  to clear the tag for the next thread.


  -> Why we don't need it here indeed ?

  There are 3 conditions making the processor removing the exclusive acces
  tag (c.f. datasheet section 12.5.7.2 page 78):
    - exec a CLREX instruction,
    - exec a STREX instruction wich succed or failed,
    - ANNNNND if an exception occurs...

  and cool, the task switching is done by an exception, so each time we switch
  of thread, the CLREX is automatically/transparenly done. Hence in fact, we
  don't need to use it ourselves.
 */
__attribute__ ( ( naked ) ) void pendSVHook (void)
{
  /* Push the context on the current process's stack.
     push is a synonym for STMDB sp!, relist */
  asm volatile (".save_outgoin_context: \n\t") ;
  asm volatile ("push    {r4 - r11, lr} \n\t") ;

  /* Guess the outgoing process id. */
  outgoing_proc_id = kernel.current_process_id ;
  /* Only enqueue the outcomming process if it is not the "Idle" task. */
  if (outgoing_proc_id != MAIN_PROCESS_ID)
    enqueue (&kernel.queue, outgoing_proc_id) ;
  /* Guess the incoming process id. */
  incoming_proc_id = take (&kernel.queue) ;
  /* Record the new running process. incoming_proc_id could be removed and
     replaced by kernel.current_process_id in fact. */
  kernel.current_process_id = incoming_proc_id ;

  /* Save the outgoing process' SP.
     If the process is the main process, then save its SP in the kernel
     structure, otherwise in the process's structure.
     r1 = outgoing stack pointer. */
  if (outgoing_proc_id != MAIN_PROCESS_ID) {
    outgoing_proc = &kernel.processes[outgoing_proc_id] ;
    asm volatile ("mov r1, %0     \n\t" :: "r" (&outgoing_proc->sp) : "r1") ;
  }
  else asm volatile ("mov r1, %0  \n\t" :: "r" (&kernel.main_task_sp) : "r1") ;
  /* Really save outgoing SP. */
  asm volatile ("mrs r2, msp    \n\t" ::: "r2") ;
  asm volatile ("str r2, [r1]   \n\t" ::: "memory") ;

  /* Restore the incoming process' SP.
     If the process is the main process, then load its SP from the kernel
     structure, otherwise from the process's structure.
     r0 = incoming stack pointer. */
  if (incoming_proc_id != MAIN_PROCESS_ID) {
    incoming_proc = &kernel.processes[incoming_proc_id] ;
    asm volatile ("mov r0, %0     \n\t" :: "r" (incoming_proc->sp) : "r0") ;
  }
  else asm volatile ("mov r0, %0  \n\t" :: "r" (kernel.main_task_sp) : "r0") ;
  /* Really write incoming SP. */
  asm volatile ("msr msp, r0    \n\t") ;

  /* Pop the context of the incoming process on its (the current process's)
     stack.
     pop is a synonym for LDMIA sp! reglist */
  asm volatile (".restore_incoming_context: \n\t") ;
  asm volatile ("pop {r4 - r11, lr}   \n\t" :::
		"r4", "r5", "r6", "r7","r8", "r9", "r10", "r11") ;
  /* Exit from exception by jumping at LR which must be 0xFFFFFFF9.
     See SAM3x8E datasheet 16.6.7.6 page 86. */
  asm volatile ("bx lr                       \n\t") ;
}

#ifdef __cplusplus
}  /* End of extern "C" for svcHook (). */
#endif



#ifdef __cplusplus
extern "C" {
#endif

/** Handler for the sysTick interrupt. It fires a PendSV exception. This
    exception will be processed by pendSVHook to swap processes.

    ATTENTION: must be extern "C" otherwise ignored by the linker, hence
    never called (no effective overriding).
*/
int sysTickHook ()
{
  SCB->ICSR |= SCB_ICSR_PENDSVSET_Msk ;
  return (0) ;
}

#ifdef __cplusplus
} /* End of extern "C" for sysTickHook (). */
#endif
